/*C* $Id: game.cc,v 1.10 1997/08/13 15:44:53 james Exp $
 *
 * Hatman - The Game of Kings
 * Copyright (C) 1997 James Pharaoh & Timothy Fisken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *C*/

/*
 * game.cc
 *
 * This beastie /is/ hatman. The game itself is mostly all in here. The game
 * object manages all game-related variables etc, and it's member functions
 * manage the whole lot.
 *
 * Basically, it is like this: The object is initialised and play() is called,
 * with the filename of the level to be played. play() sets everything up and
 * then calls mainLoop, which is just a big loop. mainLoop() sets quit to
 * false, and exits when it is true.
 *
 * The loop in mainLoop() first checks for keyboard input, and for arrow keys
 * sets the appropriate directions in the player object. The functions
 * moveGhosts(), movePlayer() and moveFruit() are then called, which handle the
 * movement and collisions of the named objects (movePlayer() does not handle
 * any collisions). Finally, mainLoop() keeps track of time, and if there is
 * plenty spare it redraws the screen by calling draw().
 *
 * The Game object is also responsible for making sure no-one bangs into walls,
 * and controls the movement of ghosts and fruit.
 *
 * Game also draws the score and lives indicator (drawScoreBox()).
 *
 * Addition: Game::setup() must now be called prior to each "game". play()
 * returns true if the player completed the level.
 *
 */

#include "./animation.h"
#include "./background.h"
#include "./blocks.h"
#include "./game.h"
#include "./gluetext.h"
#include "./data.h"
#include "./files.h"
#include "./fruit.h"
#include "./level.h"
#include "./misc.h"
#include "./options.h"
#include "./params.h"
#include "./player.h"
#include "./pos.h"
#include "./screen.h"
#include "./sound.h"
#include "./speedcontrol.h"
#include "./sprites.h"
#include "./ticker.h"
#include "util/util.h"
#include "vga++/font.h"
#include "vga++/keycodes.h"
#include "vga++/keys.h"
#include "vga++/sprite.h"
#include "vga++/vga++blur.h"

Game* game;

//-----------------------------------------------------------------------------
Game::Game()
{
 ghosts = new Collection<Ghost>(8, 8);
 level = new Level;
 player = new Player;
 fruit = new Fruit;
 fruitStarts = new Collection<Pos>(8, 8);
 gtm = new GlueTextManager(fontThin);
 fontThin->setHAlign(Font::Centre);
 fontThin->setVAlign(Font::Middle);
 playerSpeed = new SpeedControl;
 ghostSpeed = new SpeedControl;
 ghostScaredSpeed = new SpeedControl;
 ghostEyesSpeed = new SpeedControl;
 fruitSpeed = new SpeedControl;
}

//-----------------------------------------------------------------------------
Game::~Game()
{
 delete ghosts;
 delete level;
 delete player;
 delete fruit;
 delete fruitStarts;
 delete gtm;
 delete playerSpeed;
 delete ghostSpeed;
 delete ghostScaredSpeed;
 delete ghostEyesSpeed;
 delete fruitSpeed;
}

//-----------------------------------------------------------------------------
void Game::setup()
{
 player->reset();
}

//-----------------------------------------------------------------------------
bool Game::playerCanMove(Dir dir)
{
 Pos pos = player->getPos();
 if(pos.x == lvlW && (dir == Up || dir == Down)) return false;
 if(pos.y == lvlH && (dir == Left || dir == Right)) return false;
 pos += Pos(xInc[dir], yInc[dir]);
 if(pos.x < 0 || pos.x >= lvlW || pos.y < 0 || pos.y >= lvlH) return true;
 int sq = level->getSquare(pos);
 return (BLOCK_IS_WALL(sq) || BLOCK_IS_SCORE(sq) || BLOCK_IS_BASE(sq))? false : true;
}

//-----------------------------------------------------------------------------
bool Game::ghostCanMove(int i, Dir dir)
{
 Pos pos = (*ghosts)[i]->getPos();
 if(pos.x == lvlW && (dir == Up || dir == Down)) return false;
 if(pos.y == lvlH && (dir == Left || dir == Right)) return false;
 pos += Pos(xInc[dir], yInc[dir]);
 if(pos.x < 0 || pos.x >= lvlW || pos.y < 0 || pos.y >= lvlH) return true;
 int sq = level->getSquare(pos);
 return (BLOCK_IS_WALL(sq) || BLOCK_IS_SCORE(sq))? false : true;
}

//-----------------------------------------------------------------------------
bool Game::fruitCanMove(Dir dir)
     /*
      * Returns true if the fruit can move in the given direction.
      */
{
 Pos pos = fruit->pos + Pos(xInc[dir], yInc[dir]);
 if(pos.x < 0 || pos.x >= lvlW || pos.y < 0 || pos.y >= lvlH) return true;
 int sq = level->getSquare(pos);
 return (BLOCK_IS_WALL(sq) || BLOCK_IS_SCORE(sq) || BLOCK_IS_BASE(sq))? false : true;
}

//-----------------------------------------------------------------------------
void Game::draw()
     /*
      * This is called once a frame to update the screen. The first drawn goes
      * lowest, etc...
      */
{
 if(fruitActive) fruit->draw(); // fruit
 player->draw();                // player
 every(*ghosts, draw());        // ghosts
 gtm->blur(screen);             // gluetext
 screen->blur();
}

//-----------------------------------------------------------------------------
void Game::drawScoreBox()
     /*
      * Redraws the score box, including lives.
      */
{
 // first redraw the lives indicator
 screen->restore(lifePos.x, lifePos.y, sprW + 10 * player->getLives(), sprH);
 for(int i=0; i+1<player->getLives(); i++) // yes, one less than the total
  (*sprites)[SPRITE_EXTRA_0_0]->draw(screen, lifePos.x + 10 * i, lifePos.y);
 screen->update(lifePos.x, lifePos.y, sprW + 10 * player->getLives(), sprH);

 // then draw the score
 defaultFont.masked();
 defaultFont.color(makeColor(5, 5, 5));
 defaultFont.setHAlign(Font::Right);
 defaultFont.restoreAndWritefAndUpdate(screen, scorePos.x, scorePos.y, "%d", player->getScore());
}

//-----------------------------------------------------------------------------
void Game::kill()
     /*
      * Called when the player loses a life.
      */
{
 if(--player->lives == 0) { quit = true; return; }
 power = 0;
 player->setPos(player->startPos);
 player->move(Right);
 every(*ghosts, restart());
 drawScoreBox();
}

//-----------------------------------------------------------------------------
void Game::moveGhosts()
     /*
      * This procedure performs once-per-frame operations on every ghost in turn.
      */
{
 int aliveMoves = ghostSpeed->moves();
 int scaredMoves = ghostScaredSpeed->moves();
 int deadMoves = ghostEyesSpeed->moves();
 foreach(*ghosts, i)
  {
   Ghost* ghost = (*ghosts)[i];
   int moves = (ghost->state == Ghost::Alive)? aliveMoves : ((ghost->state == Ghost::Scared)? scaredMoves : deadMoves);
   if(moves)
    {
     if(ghost->move())
      {
       // Setup some useful variables
       Pos g = ghost->getPos(), target;
       target = (ghost->state == Ghost::Dead)? base : player->getPos();
       int xDif = abs(target.x - g.x), yDif = abs(target.y - g.y);
       Dir oldDir = ghost->getDir();
       
       // Decide which way we really want to go
       Dir dir1 = ((xDif > yDif)? ((g.x>target.x)?Left:Right) : ((g.y>target.y)?Up:Down) );
       Dir dir2 = ((xDif <= yDif)? ((g.x>target.x)?Left:Right) : ((g.y>target.y)?Up:Down) );
       if(ghost->state == Ghost::Scared) { dir1 = behindOf[dir1]; dir2 = behindOf[dir2]; }
       
       // Every now and then go in a silly direction
       if(rInt(2))
	{
	 dir1 = (Dir) rInt(4);
	 if(dir1 == Left || dir1 == Right) dir2 = rInt(1)? Up : Down;
	 else dir2 = rInt(1)? Left : Right;
	}
       
       // Decide which way to actually go
       Dir dir3 = behindOf[dir2];
       Dir dir4 = behindOf[dir1];
       Dir badDir = behindOf[oldDir];
       if (ghostCanMove(i, dir1) && dir1 != badDir) ghost->move(dir1);
       else if (ghostCanMove(i, dir2) && dir2 != badDir) ghost->move(dir2);
       else if (ghostCanMove(i, dir3) && dir3 != badDir) ghost->move(dir3);
       else if (ghostCanMove(i, dir4) && dir4 != badDir) ghost->move(dir4);
       else ghost->move(badDir);
      } // if(ghost->move())

     // check if we have collided with the player
     if(ghost->getScreenPos().near(player->getScreenPos(), Threshold))
      {
       if(ghost->state == Ghost::Scared)
	{
	 ghost->state = Ghost::Dead;
	 player->score += (ghostScore <<= 1);
	 gtm->add(new GlueText(Pos(ghost->hX() + xOff + sprW / 2, ghost->hY() + yOff + sprH / 2), scoreFlashTime, "%d",
			       ghostScore));
	 drawScoreBox();
	}
       else if (ghost->state == Ghost::Alive) kill();
       continue;
      }
    } // if(moves)

   // if we are on the base square, give us our life back
   if((ghost->getPos() == base) && (ghost->state == Ghost::Dead)) ghost->state = Ghost::Alive;
  }
}

//-----------------------------------------------------------------------------
void Game::checkInput()
{
 // take keyboard input
 if(keyboard->update())
  {
   Dir dir = None;
   
   KP_Q quit = true;
   KP_ESCAPE quit = true;
   KP_S screen->screenShot();
   KP_UP dir = Up;
   KP_DOWN dir = Down;
   KP_LEFT dir = Left;
   KP_RIGHT dir = Right;
   
   // handle direction keys
   if((dir != None) && ((dir != player->getDir()) || !player->isMoving()))
    {
     if(dir == behindOf[player->getDir()])
      {
       player->turnBack();
       player->setDirPending(None);
      }
     else
      {
       player->setDirPending(dir);
       if(!player->onSquare() && (player->getDisplacement() <= DirGuessThreshold) && playerCanMove(dir)) player->turnBack();
      }
    }
  } /* if(Keyboard->Update()) */
}

//-----------------------------------------------------------------------------
void Game::movePlayer()
{
 // do stuff when player is on a square
 if(player->onSquare())
  {
   // eat dots
   if(BLOCK_IS_DOT(level->getSquare(player->getPos())))
    {
     player->score += 10;
     drawScoreBox();
     level->clear(player->getPos());
#ifdef SOUND
     if(options->sound) Effects->Play(sndDot);
#endif
    }
   // eat pellets
   if(BLOCK_IS_PELLET(level->getSquare(player->getPos())))
    {
     player->score += 100;
     drawScoreBox();
     power = levelMode->ghostScaredTime;
     ghostScore = 50; // half score for first ghost
     every(*ghosts, pellet());
     Ghost::flash = 0;
     every(*ghosts, turnBack());
     level->clear(player->getPos());
    }
   // change direction if requested
   if(player->getDirPending() != None && playerCanMove(player->getDirPending())) player->move(player->getDirPending());
   else if(!playerCanMove(player->getDir())) player->stop();
  }

 if(playerSpeed->moves())
  {
   player->move();
   checkInput();
  }
 
 // control power pellets
 if(power)
  {
   if(--power == 0)
    {
     every(*ghosts, pelletEnd());
     Ghost::flash = 0;
    }
   else if(power < ghostFlashes * ghostFlashTime)
    Ghost::flash = ((power % (ghostFlashTime * 2)) / ghostFlashTime)? 1 : 2;
  }

 animPlayer->advanceSubFrame();
}

//-----------------------------------------------------------------------------
void Game::moveFruit()
     /*
      * Called once per frame to handle all things fruit-related
      */
{
 if(fruitActive)
  {
   if(fruit->getScreenPos().near(player->getScreenPos(), Threshold))
    {
     fruitActive = false;
     player->score += levelMode->fruitScore;
     gtm->add(new GlueText(fruit->getScreenPos() + Pos(sprW / 2, sprH / 2), scoreFlashTime, "%d", levelMode->fruitScore));
     drawScoreBox();
    }
   else if(fruitSpeed->moves() && fruit->move())
    {
     // the fruit has moved 1 square...
     // if we are off the screen disappear
     if(fruit->pos.x == -1 || fruit->pos.y == -1 || fruit->pos.x == lvlW || fruit->pos.y == lvlH)
      fruitActive = false;
     else
      {
       // otherwse, figure out where to go next
       Dir badDir = behindOf[fruit->dir];
       Dir dir1 = rInt(3)? fruitDir : (Dir)rInt(4); // 1 in 3 times go randomly
       Dir dir2 = rInt(2)? leftOf[dir1] : rightOf[dir1];
       Dir dir3 = behindOf[dir2];
       Dir dir4 = behindOf[dir1];
       if(dir1 != badDir && fruitCanMove(dir1)) fruit->dir = dir1;
       else if(dir2 != badDir && fruitCanMove(dir2)) fruit->dir = dir2;
       else if(dir3 != badDir && fruitCanMove(dir3)) fruit->dir = dir3;
       else if(dir4 != badDir && fruitCanMove(dir4)) fruit->dir = dir4;
       else fruit->dir = badDir;
      }
    }
  }

 // if there is no fruit, make some randomly
 else if(!rInt(levelMode->fruitChance) && (*fruitStarts) > 0)
  {
   fruitActive = true;
   
   // pick a starting position at random
   Pos pos = *((*fruitStarts)[rInt(*fruitStarts)]);
   fruit->pos = pos;
   
   // figure out what direction we should be going
   if(pos.x == -1) fruitDir = Right;
   else if(pos.y == -1) fruitDir = Down;
   else if(pos.x == lvlW) fruitDir = Left;
   else if(pos.y == lvlH) fruitDir = Up;
   fruit->dir = fruitDir;

   // set the type. needs work.
   fruit->type = 0;
  }
}

//-----------------------------------------------------------------------------
void Game::mainLoop()
{
 // MAIN GAME LOOP
 int frames = 0, ticks = 0;
 Ticker ticker(1);
 quit = false;
 while(!quit)
  {
   moveGhosts();
   movePlayer();
   moveFruit();

   gtm->tick();

   // if we eat all the dots, the level is finished
   if(level->dots == 0) quit = true;

   // update the screen and yield if there is sufficient time
   ticker.tick(); ticks++;
   if(ticker.spare()) { draw(); ticker.yield(); frames++; }
  }
#ifndef NDEBUG
 printf("%d f/s\n", frames * 1000 / ticks);
#endif
}

//-----------------------------------------------------------------------------
bool Game::play()
{
 // check the level
 if(!level->checkLevel()) { nonFatal("Invalid level"); return false; }

 // initialise
 srand(time(0));
 power = 0;
 xOff = sprW / 2;
 yOff = sprH / 2;

 // play initial sound effect
#ifdef SOUND
 if(options->sound) Effects->Play(sndBegin, 0);
#endif
 
 // take object positions from the level
 player->startPos = level->find(BLOCK_START);
 base = level->find(BLOCK_BASE);
 scorePos = level->find(BLOCK_SCORE_2).fromLevel() + Pos(sprW - (lvlW - 8) / 2, (lvlH - 8) / 2);
 lifePos = level->find(BLOCK_SCORE_0).fromLevel();
 level->clear(player->startPos);
 level->getGhosts(ghosts);

 // setup fruit and stuff
 fruitStarts->delAll();
 while(1)
  {
   Pos pos = level->find(BLOCK_FRUIT);
   if(pos.x == -1) break;
        if(pos.x == 0) fruitStarts->add(new Pos(-1, pos.y));
   else if(pos.y == 0) fruitStarts->add(new Pos(pos.x, -1));
   else if(pos.x == lvlW-1) fruitStarts->add(new Pos(lvlW, pos.y));
   else if(pos.y == lvlH-1) fruitStarts->add(new Pos(pos.x, lvlH));
   level->clear(pos);
  }
 fruitActive = false;
 fruit->type = levelMode->fruit;

 // initialise the player
 player->setPos(player->startPos);
 player->move(Right);
 animPlayer->setRepeats(levelMode->mouthDelay);

 // set up speeds
 *playerSpeed = levelMode->playerSpeed;
 *ghostSpeed = levelMode->ghostSpeed;
 *ghostScaredSpeed = levelMode->ghostScaredSpeed;
 *ghostEyesSpeed = levelMode->ghostEyesSpeed;
 *fruitSpeed = levelMode->fruitSpeed;
 
 // initialise the keyboard
 keyboard->clear();
 keyboard->setMode(0);
 
 // initlialise the screen
 bk->copyScreen(*screen);
 level->draw();
 screen->update();
 drawScoreBox();
 screen->enableClipping();

 mainLoop();

 // clean up
 ghosts->delAll();

 return (level->dots == 0)? true : false;
}

//-----------------------------------------------------------------------------
bool Game::play(const char* filename)
{
 if(!level->load(libFileName(filename))) return false;
 levelMode = defaultLevelMode;
 return play();
}

//-----------------------------------------------------------------------------
bool Game::play(const Level* _level)
{
 *level = *_level;
 levelMode = defaultLevelMode;
 return play();
}

//-----------------------------------------------------------------------------
bool Game::play(LevelMode* _levelMode)
{
 if(!level->load(libFileName(_levelMode->filename))) return false;
 levelMode = _levelMode;
 return play();
}
