/*C* $Id: font.cc,v 1.2 1997/08/11 19:22:18 james Exp $
 *
 * Hatman - The Game of Kings
 * Copyright (C) 1997 James Pharaoh & Timothy Fisken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *C*/

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "util/error.h"
#include "util/file.h"
#include "util/util.h"
#include "vga++/vga++.h"
#include "vga++/vga++blur.h"
#include "vga++/font.h"

//-----------------------------------------------------------------------------
Font defaultFont(gl_font8x8, 8, 8, false);

//----------------------------------------------------------------------------- class Font
Font* Font::current;

//-----------------------------------------------------------------------------
Font::Font()
{
 data = NULL;
 fg = 1; bg = 0;
 hAlign = Left;
 vAlign = Top;
 wm = WRITEMODE_OVERWRITE;
}

//-----------------------------------------------------------------------------
Font::Font(unsigned char* _data, int _w, int _h, bool expanded, int _fg)
{
 w = _w; h = _h;
 fg = _fg; bg = 0;
 if(expanded)
  {
   data = new unsigned char [w * h * 256];
   memcpy(data, _data, w * h * 256);
   wm = WRITEMODE_OVERWRITE | FONT_EXPANDED;
  }
 else
  {
   int size = (int)((w + 7) >> 3) * h * 256;
   data = new unsigned char [size];
   memcpy(data, _data, size);
   wm = WRITEMODE_OVERWRITE | FONT_COMPRESSED;
  }
 current = NULL;
 hAlign = Left;
 vAlign = Top;
}

//-----------------------------------------------------------------------------
Font::~Font()
{
 if(data) delete data; 
}

//-----------------------------------------------------------------------------
bool Font::load(const char* filename)
{
 current = NULL; // prolly best if this goes at the start, eh?

 File* f = gzFilename(filename)? new GzipPipe(filename, "r") : new File(filename, "r");
 if(!*f) goto error;
 
 f->read(&w, sizeof(w));
 f->read(&h, sizeof(h));
 if(f->error()) goto error;
 
 if((w > 0xFFFF) || (h > 0xFFFF))
  {
   if(int stat = f->close()) setError("%s: child returned status: %d\n", filename, stat);
   else setError("%s: File Invalid", filename);
   return false;
  }
 
 if(data) delete data;
 data = new unsigned char[w * h * 256];
 
 if(!f->read(data, w * h * 256)) goto errorDelete;

 if(f->close()) goto errorDelete;

 return true;

errorDelete:
 delete data;
 data = NULL;
error:
 if(int stat = f->close()) setError("%s: child returned status: %d\n", filename, stat);
 else setError("%s: %y", filename);
 return false;
}

//-----------------------------------------------------------------------------
void Font::select()
{
 if(current != this)
  {
   current = this;
   gl_setfont(w, h, data);
   gl_setwritemode(wm);
   gl_setfontcolors(bg, fg);
  }
}

//-----------------------------------------------------------------------------
void Font::restoreAndWritefAndUpdate(BlurryVga* vc, int x, int y, const char* format, ...)
{
 if(!data) return;
 va_list ap; va_start(ap, format);
 char *s;
 vasprintf(&s, format, ap);
 va_end(ap);
 int _w = width(s);
 alignCoords(x, y, _w);
 vc->restore(x, y, _w, h);
 write(vc, x, y, s);
 vc->update(x, y, _w, h);
 free(s);
}

//-----------------------------------------------------------------------------
void Font::writef(VgaContext* vc, int x, int y, const char* format, ...)
{
 if(!data) return;
 va_list ap; va_start(ap, format);
 char *s;
 vasprintf(&s, format, ap);
 va_end(ap);
 int _w = width(s);
 alignCoords(x, y, _w);
 write(vc, x, y, s);
 free(s);
}

//-----------------------------------------------------------------------------
void Font::writefAndBlur(BlurryVga* vc, int x, int y, const char* format, ...)
{
 if(!data) return;
 va_list ap; va_start(ap, format);
 char* s;
 vasprintf(&s, format, ap);
 va_end(ap);
 int _w = width(s);
 alignCoords(x, y, _w);
 write(vc, x, y, s);
 vc->blur(x, y, _w, h);
 free(s);
}

//-----------------------------------------------------------------------------
void Font::writefAndUpdate(BlurryVga* vc, int x, int y, const char* format, ...)
{
 if(!data) return;
 va_list ap; va_start(ap, format);
 char *s;
 vasprintf(&s, format, ap);
 va_end(ap);
 int _w = width(s);
 alignCoords(x, y, _w);
 write(vc, x, y, s);
 vc->update(x, y, _w, h);
 free(s);
}

//-----------------------------------------------------------------------------
int Font::width(const char* s)
{
 return w * strlen(s);
}

//-----------------------------------------------------------------------------
int Font::widthf(const char* format, ...)
{
 va_list ap; va_start(ap, format);
 char* s;
 vasprintf(&s, format, ap);
 va_end(ap);
 int len = width(s);
 free(s);
 return len;
}

//-----------------------------------------------------------------------------
void Font::alignCoords(int& x, int& y, int l)
{
 switch(hAlign)
  {
  case Left:
   break;
  case Centre:
   x -= l/2;
   break;
  case Right:
   x -= l;
   break;
  }
 switch(vAlign)
  {
  case Top:
   break;
  case Middle:
   y -= h/2;
   break;
  case Bottom:
   y -= h;
   break;
  }
}

//-----------------------------------------------------------------------------
void Font::write(class VgaContext* vc, int x, int y, const char* s)
{
 vc->select();
 select();
 gl_write(x, y, (char*)s);
}

//----------------------------------------------------------------------------- class PropFont
PropFont::PropFont(unsigned char* _data, int _w, int _h, bool expanded)
 : Font(_data, _w, _h, expanded)
{
 calcWidths();
}

//-----------------------------------------------------------------------------
bool PropFont::load(const char* filename)
{
 bool ret = Font::load(filename);
 if(ret) calcWidths(); // this gets cross if there is no data :)
 return ret;
}

//-----------------------------------------------------------------------------
int PropFont::width(const char* s)
{
 int len = 0;
 for(; *s; s++)
  len += charWidths[*s];
 return len;
}

//-----------------------------------------------------------------------------
void PropFont::write(class VgaContext* vc, int _x, int _y, const char* s)
{
 vc->select();
 for(const char* p=s; *p; p++)
  {
   unsigned char* q = &data[w * h * (*p)];
   gl_putboxmask(_x, _y, w, h, q);
   _x += charWidths[*p];
  }
}

//-----------------------------------------------------------------------------
void PropFont::calcWidths()
{
 unsigned char* p = data;
 for(int i=0; i<256; i++)
  {
   charWidths[i] = 0;
   for(int x=0; x<w; x++)
    {
     bool dirty = false;
     for(int y=0; y<h; y++)
      if(p[y * w + x]) { dirty = true; break; }
     if(dirty) charWidths[i] = x + 1;
    }
   p += w * h;
  }
 // empty the space character
 p = &data[w * h * 32];
 for(int i=0; i<w*h; i++) *p++ = 0;
}


