/*C* $Id: sprite.cc,v 1.2 1997/08/11 19:22:18 james Exp $
 *
 * Hatman - The Game of Kings
 * Copyright (C) 1997 James Pharaoh & Timothy Fisken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *C*/

#include "vga++/sprite.h"
#include "util/error.h"
#include "util/file.h"
#include "util/util.h"
#include "vga++/vga++.h"
#include "vga++/vga++blur.h"

//-----------------------------------------------------------------------------
Sprite::Sprite()
{
 ok = true;
 data = NULL;
 compiledData = NULL;
}

//-----------------------------------------------------------------------------
Sprite::Sprite(int _w, int _h, FILE *fp)
{
 ok = true;
 data = NULL;
 compiledData = NULL;
 read(fp, _w, _h);
}

//-----------------------------------------------------------------------------
Sprite::Sprite(int _w, int _h, int Initial = 0)
{
 ok = true;
 data = NULL;
 compiledData = NULL;
 w = _w; h = _h;
 alloc();
 for(int x=0; x<w; x++)
  for(int y=0; y<h; y++)
   data[y][x] = Initial;
}

//-----------------------------------------------------------------------------
Sprite::Sprite(int _w, int _h, const char* _data)
{
 ok = true;
 data = NULL;
 compiledData = NULL;
 w = _w; h = _h;
 alloc();
 for(int y=0; y<h; y++)
  for(int x=0; x<w; x++)
   data[y][x] = *_data++;
}

//-----------------------------------------------------------------------------
Sprite::~Sprite()
{
 free();
}

//-----------------------------------------------------------------------------
void Sprite::alloc(char**& _data)
{
 char* realData = new char[h*w];
 _data = new char*[h];
 for(int i=0; i<h; i++) _data[i] = realData + w*i; 
}

//-----------------------------------------------------------------------------
void Sprite::alloc()
{
 free();
 alloc(data);
 ok = true;
}

//-----------------------------------------------------------------------------
void Sprite::free()
{
 if(!ok) return;
 freeCompiled();
 if(!data) return;
 // evil, nasty and Bill-Gates-ish, but sadly correct!
 delete data[0];
 delete data;
}

void Sprite::freeCompiled()
{
 if(compiledData) delete compiledData;
 compiledData = NULL;
}

//-----------------------------------------------------------------------------
bool Sprite::read(FILE* fp, int _w, int _h)
{
 w = _w; h = _h;
 alloc();
 if(!fread(data[0], w * h, 1, fp))
  {
   setError("cannot read data: %y");
   return ok = false;
  }
 return ok = true;
}

//-----------------------------------------------------------------------------
bool Sprite::load(const char* filename)
{
 File* f = new File(filename, "r");
 if(!*f)
  {
   setError("cannot open %s: %y", filename);
   return ok = false;
  }
 f->read(&w, sizeof(w));
 f->read(&h, sizeof(h));
 if(f->error())
  {
   setError("%s: cannot read header: %y");
   delete f;
   return ok = false;
  }
 if(!read(*f, w, h))
  {
   setError("cannot load %s: %z", filename);
   delete f;
   return ok = false;
  }
 if(f->close())
  {
   setError("cannot close %s: %y", filename);
   delete f;
   return ok = false;
  }
 delete f;
 return true;
}

//-----------------------------------------------------------------------------
bool Sprite::write(FILE* fp)
{
 if(!ok || !data) return false;
 if(!fwrite(data[0], w * h, 1, fp))
  {
   setError("cannot write data: %y");
   return false;
  }
 return true;
}

//-----------------------------------------------------------------------------
bool Sprite::save(const char* filename)
{
 if(!ok || !data) return false;
 FILE* fp = fopen(filename, "w");
 if(!fp) return false;
 fwrite(&w, sizeof(w), 1, fp);
 fwrite(&h, sizeof(h), 1, fp);
 if(ferror(fp)) { fclose(fp); return false; }
 bool ret = write(fp);
 fclose(fp);
 return ret;
}

//-----------------------------------------------------------------------------
void Sprite::flipH()
{
 if(!ok || !data) return;
 freeCompiled();
 for(int x=0; x<w/2; x++) for(int y=0; y<h; y++)
  Swap(data[y][x], data[y][w-x-1]);
}

//-----------------------------------------------------------------------------
void Sprite::flipV()
{
 if(!ok || !data) return;
 freeCompiled();
 for(int y=0; y<h/2; y++) for(int x=0; x<w; x++)
  Swap(data[y][x], data[h-y-1][x]);
}

//-----------------------------------------------------------------------------
void Sprite::rotate()
{
 if(!ok || !data) return;
 freeCompiled();
 char** temp;
 alloc(temp);
 for(int x = 0; x < w; x++) for(int y = 0; y < h; y++)
  temp[y][x] = data[h-x - 1][y];
 free();
 data = temp;
}

//-----------------------------------------------------------------------------
void Sprite::compile()
{
 if(!ok || !data) return;
 freeCompiled();
 compiledData = new char[gl_compiledboxmasksize(w, h, data[0])];
 gl_compileboxmask(w, h, data[0], compiledData);
}

//-----------------------------------------------------------------------------
void Sprite::compileAndForget()
{
 if(!ok || !data) return;
 compile();
 delete data[0]; // Yes indeed
 delete data;
}

//-----------------------------------------------------------------------------
struct ClippedCoords
{
 int x, y, w, h;
};

//-----------------------------------------------------------------------------
void Sprite::draw(VgaContext* vc, int x, int y, ClippedCoords* cc)
{
 if(!ok) return;
 if(vc->isClipping() && ((x<0) || (y<0) || ((x+w) > vgaW) || ((y+h) > vgaH)))
  {
   int nx = x, ny = y, nw = w, nh = h;
   if(nx < 0) { nw -= -nx; nx = 0; }
   if(ny < 0) { nh -= -ny; ny = 0; }
   if((nx + nw) > vgaW) { nw -= (nx + nw) - vgaW; }
   if((ny + nh) > vgaH) { nh -= (ny + nh) - vgaH; }
   if(cc != NULL) { cc->x = nx; cc->y = ny; cc->w = nw; cc->h = nh; }
   if(data)
    {
     vc->select();
     char* dst = VBUF + ny * BYTEWIDTH + nx;
     for(int cy = 0; cy<nh; cy++)
      {
       char* src = data[ny - y + cy] + nx - x;
       for(int cx = 0; cx<nw; cx++)
	{
	 if(*src) *dst = *src;
	 dst++; src++;
	}
       dst += BYTEWIDTH - nw;
      }
    }
  }
 else
  {
   if(compiledData) vc->putBoxMaskCompiled(x, y, w, h, compiledData);
   else if(data) vc->putBoxMask(x, y, w, h, data[0]);
  }
}

//-----------------------------------------------------------------------------
void Sprite::blur(BlurryVga* vc, int x, int y)
{
 if(!ok) return;
 ClippedCoords cc;
 cc.x = x; cc.y = y; cc.w = w; cc.h = h;
 draw(vc, x, y, &cc);
 vc->blur(cc.x, cc.y, cc.w, cc.h);
}

//-----------------------------------------------------------------------------
void Sprite::keep(BlurryVga* vc, int x, int y)
{
 if(!ok) return;
 ClippedCoords cc;
 cc.x = x; cc.y = y; cc.w = w; cc.h = h;
 draw(vc, x, y, &cc);
 vc->keep(cc.x, cc.y, cc.w, cc.h);
}

//-----------------------------------------------------------------------------
void Sprite::drawAndUpdate(BlurryVga* vc, int x, int y)
{
 if(!ok) return;
 ClippedCoords cc;
 cc.x = x; cc.y = y; cc.w = w; cc.h = h;
 draw(vc, x, y, &cc);
 vc->update(cc.x, cc.y, cc.w, cc.h);
}

//-----------------------------------------------------------------------------
void Sprite::restoreAndDrawAndUpdate(BlurryVga* vc, int x, int y)
{
 if(!ok) return;
 ClippedCoords cc;
 cc.x = x; cc.y = y; cc.w = w; cc.h = h;
 vc->restore(cc.x, cc.y, cc.w, cc.h);
 draw(vc, x, y, &cc);
 vc->update(cc.x, cc.y, cc.w, cc.h);
}

//-----------------------------------------------------------------------------
char Sprite::setPixel(int x, int y, int c)
{
 if(!ok || !data) return c;
 freeCompiled();
 return data[y][x] = c;
}
